import subprocess

import typer_cloup as typer
from docs_src.parameter_types.enum import tutorial001 as mod
from typer_cloup.testing import CliRunner

runner = CliRunner()

app = typer.Typer()
app.command()(mod.main)


def test_help():
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "--network [simple|conv|lstm]" in result.output


def test_main():
    result = runner.invoke(app, ["--network", "conv"])
    assert result.exit_code == 0
    assert "Training neural network of type: conv" in result.output


def test_invalid():
    result = runner.invoke(app, ["--network", "capsule"])
    assert result.exit_code != 0
    assert (
        "Error: Invalid value for '--network': 'capsule' is not one of 'simple', 'conv', 'lstm'"
        in result.output
    )


def test_script():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, "--help"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
    )
    assert "Usage" in result.stdout



import typer_cloup as typer


def main(in_prod: bool = typer.Option(True, " /--demo", " /-d")):
    if in_prod:
        typer.echo("Running in production")
    else:
        typer.echo("Running demo")


if __name__ == "__main__":
    typer.run(main)



from typing import Optional

import pytest
from cloup import OptionGroup

import typer_cloup as typer
from typer_cloup import option_group
from typer_cloup.testing import CliRunner

runner = CliRunner()


def test_no_option():
    app = typer.Typer()

    with pytest.raises(ValueError) as exc_info:

        @app.command()
        @option_group("Some group")
        def main(foo: Optional[str] = typer.Option(None)):  # pragma: no cover
            pass

    assert exc_info.value.args[0] == "you must provide at least one option"


def test_invalid_group_name():
    app = typer.Typer()

    with pytest.raises(TypeError) as exc_info:

        @app.command()
        @option_group(None)
        def main(foo: Optional[str] = typer.Option(None)):  # pragma: no cover
            pass

    assert (
        exc_info.value.args[0]
        == "the first argument of `@option_group` must be its title, a string; you probably forgot it"
    )


def test_not_parameter():
    app = typer.Typer()

    @app.command()
    @option_group("Some group", "bar")
    def main(foo: Optional[str] = typer.Option(None)):
        pass  # pragma: no cover

    with pytest.raises(ValueError) as exc_info:
        runner.invoke(app, ["--foo", "foo"])
    assert exc_info.value.args[0] == "`bar` is not a parameter"


def test_not_option():
    app = typer.Typer()

    @app.command()
    @option_group("Some group", "foo")
    def main(foo: str = typer.Argument(...)):
        pass  # pragma: no cover

    with pytest.raises(TypeError) as exc_info:
        runner.invoke(app, ["--foo", "foo"])
    assert (
        exc_info.value.args[0]
        == "only parameter of type `Option` can be added to option groups"
    )


def test_double_assignment():
    app = typer.Typer()

    another_group = OptionGroup("Another group")

    @app.command()
    @option_group("Some group", "foo")
    def main(
        foo: str = typer.Option(None, group=another_group),
    ):
        pass  # pragma: no cover

    with pytest.raises(ValueError) as exc_info:
        runner.invoke(app, ["--foo", "foo"])
    assert (
        exc_info.value.args[0]
        == "`<TyperOption foo>` was first assigned to `OptionGroup('Another group', options=[])` and then passed as argument to `@option_group('Some group', ...)`"
    )



import os
import subprocess

from docs_src.commands.index import tutorial002 as mod


def test_completion_complete_subcommand_zsh():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, " "],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TUTORIAL002.PY_COMPLETE": "zsh_complete",
            "COMP_WORDS": "tutorial002.py ",
            "COMP_CWORD": "1",
            "_TYPER_COMPLETE_TESTING": "True",
        },
    )
    assert "plain\ncreate\n_\n" "plain\ndelete\n_\n" in result.stdout


def test_completion_complete_subcommand_fish():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, " "],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TUTORIAL002.PY_COMPLETE": "fish_complete",
            "COMP_WORDS": "tutorial002.py ",
            "COMP_CWORD": "",
            "_TYPER_COMPLETE_TESTING": "True",
        },
    )
    assert "plain,create\n" "plain,delete\n" in result.stdout



import os
import subprocess
from pathlib import Path
from unittest import mock

import click
import pytest
import shellingham

import typer_cloup as typer
from typer_cloup.main import solve_typer_info_defaults, solve_typer_info_help
from typer_cloup.models import TyperInfo
from typer_cloup.testing import CliRunner

runner = CliRunner()


def test_help_from_info():
    # Mainly for coverage/completeness
    value = solve_typer_info_help(TyperInfo())
    assert value is None


def test_defaults_from_info():
    # Mainly for coverage/completeness
    value = solve_typer_info_defaults(TyperInfo())
    assert value


def test_install_invalid_shell():
    app = typer.Typer()

    @app.command()
    def main():
        typer.echo("Hello World")

    with mock.patch.object(
        shellingham, "detect_shell", return_value=("xshell", "/usr/bin/xshell")
    ):
        result = runner.invoke(app, ["--install-completion"])
        assert "Shell 'xshell' is not supported." in result.stdout
    result = runner.invoke(app)
    assert "Hello World" in result.stdout


def test_callback_too_many_parameters():
    app = typer.Typer()

    def name_callback(ctx, param, val1, val2):
        pass  # pragma: no cover

    @app.command()
    def main(name: str = typer.Option(..., callback=name_callback)):
        pass  # pragma: no cover

    with pytest.raises(click.ClickException) as exc_info:
        runner.invoke(app, ["--name", "Camila"])
    assert (
        exc_info.value.message == "Too many CLI parameter callback function parameters"
    )


def test_callback_2_untyped_parameters():
    app = typer.Typer()

    def name_callback(ctx, value):
        typer.echo(f"info name is: {ctx.info_name}")
        typer.echo(f"value is: {value}")

    @app.command()
    def main(name: str = typer.Option(..., callback=name_callback)):
        typer.echo("Hello World")

    result = runner.invoke(app, ["--name", "Camila"])
    assert "info name is: main" in result.stdout
    assert "value is: Camila" in result.stdout


def test_callback_3_untyped_parameters():
    app = typer.Typer()

    def name_callback(ctx, param, value):
        typer.echo(f"info name is: {ctx.info_name}")
        typer.echo(f"param name is: {param.name}")
        typer.echo(f"value is: {value}")

    @app.command()
    def main(name: str = typer.Option(..., callback=name_callback)):
        pass

    result = runner.invoke(app, ["--name", "Camila"])
    assert "info name is: main" in result.stdout
    assert "param name is: name" in result.stdout
    assert "value is: Camila" in result.stdout


def test_completion_untyped_parameters():
    file_path = Path(__file__).parent / "assets/completion_no_types.py"
    result = subprocess.run(
        ["coverage", "run", str(file_path)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_COMPLETION_NO_TYPES.PY_COMPLETE": "zsh_complete",
            "COMP_WORDS": "completion_no_types.py --name Sebastian --name Ca",
            "COMP_CWORD": "4",
            "_TYPER_COMPLETE_TESTING": "True",
        },
    )
    assert "info name is: completion_no_types.py" in result.stderr
    assert "param is: name" in result.stderr
    assert "incomplete is: Ca" in result.stderr
    assert (
        "plain\nCamila\nThe reader of books.\n"
        "plain\nCarlos\nThe writer of scripts.\n"
        "plain\nSebastian\nThe type hints guy.\n" in result.stdout
    )

    result = subprocess.run(
        ["coverage", "run", str(file_path)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
    )
    assert "Hello World" in result.stdout


def test_completion_untyped_parameters_different_order_correct_names():
    file_path = Path(__file__).parent / "assets/completion_no_types_order.py"
    result = subprocess.run(
        ["coverage", "run", str(file_path)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_COMPLETION_NO_TYPES_ORDER.PY_COMPLETE": "zsh_complete",
            "COMP_WORDS": "completion_no_types.py --name Sebastian --name Ca",
            "COMP_CWORD": "4",
            "_TYPER_COMPLETE_TESTING": "True",
        },
    )
    assert "info name is: completion_no_types_order.py" in result.stderr
    assert "param is: name" in result.stderr
    assert "incomplete is: Ca" in result.stderr
    assert (
        "plain\nCamila\nThe reader of books.\n"
        "plain\nCarlos\nThe writer of scripts.\n" in result.stdout
    )

    result = subprocess.run(
        ["coverage", "run", str(file_path)],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
    )
    assert "Hello World" in result.stdout


def test_autocompletion_too_many_parameters():
    app = typer.Typer()

    def name_callback(ctx, param, incomplete, val2):
        pass  # pragma: no cover

    @app.command()
    def main(name: str = typer.Option(..., shell_complete=name_callback)):
        pass  # pragma: no cover

    with pytest.raises(click.ClickException) as exc_info:
        runner.invoke(app, ["--name", "Camila"])
    assert (
        exc_info.value.message == "Invalid shell-completion callback parameters: val2"
    )


def test_forward_references():
    app = typer.Typer()

    @app.command()
    def main(arg1, arg2: int, arg3: "int", arg4: bool = False, arg5: "bool" = False):
        typer.echo(f"arg1: {type(arg1)} {arg1}")
        typer.echo(f"arg2: {type(arg2)} {arg2}")
        typer.echo(f"arg3: {type(arg3)} {arg3}")
        typer.echo(f"arg4: {type(arg4)} {arg4}")
        typer.echo(f"arg5: {type(arg5)} {arg5}")

    result = runner.invoke(app, ["Hello", "2", "invalid"])

    assert (
        "Error: Invalid value for 'ARG3': 'invalid' is not a valid integer"
        in result.stdout
    )
    result = runner.invoke(app, ["Hello", "2", "3", "--arg4", "--arg5"])
    assert (
        "arg1: <class 'str'> Hello\narg2: <class 'int'> 2\narg3: <class 'int'> 3\narg4: <class 'bool'> True\narg5: <class 'bool'> True\n"
        in result.stdout
    )


def test_context_settings_inheritance_single_command():
    app = typer.Typer(context_settings=dict(help_option_names=["-h", "--help"]))

    @app.command()
    def main(name: str):
        pass  # pragma: no cover

    result = runner.invoke(app, ["main", "-h"])
    assert "Show this message and exit." in result.stdout


def test_invoke():
    app = typer.Typer()
    foo_app = typer.Typer(name="foo")

    app.add_sub(foo_app)

    @app.command()
    def invoke():
        app.invoke(hello)

    @app.command()
    def forward(name: str):
        app.forward(hello)

    @foo_app.command()
    def hello(name: str = "World"):
        typer.echo(f"Hello {name}")

    result = runner.invoke(app, ["invoke"])
    assert result.exit_code == 0
    assert "Hello World" in result.stdout

    result = runner.invoke(app, ["forward", "John"])
    assert result.exit_code == 0
    assert "Hello John" in result.stdout


def test_invoke_callback_not_command():
    app = typer.Typer()

    @app.command()
    def invoke():
        app.invoke(None)

    @app.command()
    def forward(name: str):
        app.forward(None)

    with pytest.raises(TypeError) as exc_info:
        runner.invoke(app, ["invoke"], catch_exceptions=False)
    assert exc_info.value.args[0] == "callback is not a command"

    with pytest.raises(TypeError) as exc_info:
        runner.invoke(app, ["forward", "John"], catch_exceptions=False)
    assert exc_info.value.args[0] == "callback is not a command"


def test_help_defaults():
    app = typer.Typer(context_settings=dict(auto_envvar_prefix="TEST"))

    @app.command()
    def main(
        name: str = typer.Option("John"),
        lastname: str = typer.Option("Doe", show_default="Mr. Doe"),
        age: int = typer.Option(lambda: 42, show_default=True),
    ):
        pass  # pragma: no cover

    result = runner.invoke(app, ["main", "--help"])
    assert result.exit_code == 0
    assert "John" in result.output
    assert "Mr. Doe" in result.output
    assert "(dynamic)" in result.output


def test_help_auto_envvar_prefix():
    app = typer.Typer(context_settings=dict(auto_envvar_prefix="TEST"))

    @app.command()
    def main(name: str = typer.Option(...)):
        pass  # pragma: no cover

    result = runner.invoke(app, ["main", "--help"])
    assert result.exit_code == 0
    assert "env var: TEST_NAME" in result.output


def test_help_slash_option():
    app = typer.Typer()

    @app.command()
    def main(name: str = typer.Option(..., "/name")):
        pass  # pragma: no cover

    result = runner.invoke(app, ["main", "--help"])
    assert result.exit_code == 0
    assert "/name TEXT" in result.output


def test_help_hidden_parameter():
    app = typer.Typer()

    @app.command()
    def main(
        name: str = typer.Argument(..., hidden=True),
        lastname: str = typer.Option(..., hidden=True),
    ):
        pass  # pragma: no cover

    result = runner.invoke(app, ["main", "--help"])
    assert result.exit_code == 0
    assert "  NAME" not in result.output
    assert "--lastname" not in result.output


def test_help_hidden_group():
    app = typer.Typer()

    @app.command()
    def main(
        name: str = typer.Argument(..., hidden=True),
        lastname: str = typer.Argument(..., hidden=True),
    ):
        pass  # pragma: no cover

    result = runner.invoke(app, ["main", "--help"])
    assert result.exit_code == 0
    assert "NAME" in result.output
    assert "LASTNAME" in result.output
    assert "Arguments:" not in result.output



import typer_cloup as typer


def main(name: str = typer.Argument("World", envvar="AWESOME_NAME", show_envvar=False)):
    typer.echo(f"Hello Mr. {name}")


if __name__ == "__main__":
    typer.run(main)



import typer_cloup as typer

app = typer.Typer()


@app.command()
def main(name: str, email: str = typer.Option(..., prompt=True)):
    typer.echo(f"Hello {name}, your email is: {email}")


if __name__ == "__main__":
    app()



import typer_cloup as typer

app = typer.Typer()

users_app = typer.Typer()
app.add_sub(users_app)


@users_app.callback()
def users():
    """
    Manage users in the app.
    """


@users_app.command()
def create(name: str):
    typer.echo(f"Creating user: {name}")


if __name__ == "__main__":
    app()



import typer_cloup as typer
from typer_cloup.testing import CliRunner

runner = CliRunner()

app = typer.Typer(
    context_settings=dict(
        help_option_names=["-h", "--help"],
    )
)


@app.command()
def main():
    pass  # pragma: no cover


def test_help_short_option():
    result = runner.invoke(app, ["-h"])
    assert result.exit_code == 0
    assert "Usage:" in result.output



import shlex
from typing import Any, Optional, Sequence

import click


class ShellQuotedList(click.ParamType):
    """Converts a shell-quoted string into a list of strings."""

    name = "shell-quoted list"

    def convert(
        self,
        value: Any,
        param: Optional["click.Parameter"],
        ctx: Optional["click.Context"],
    ) -> Any:
        if value is None:
            return None
        elif isinstance(value, list):
            return value

        return shlex.split(value)

    def split_envvar_value(self, rv: str) -> Sequence[str]:
        return rv  # pragma: no cover

    def __repr__(self) -> str:
        return "SHELL-QUOTED LIST"


SHELL_QUOTED_LIST = ShellQuotedList()



from typing import Optional

from cloup.constraints import RequireAtLeast, mutually_exclusive

import typer_cloup as typer
from typer_cloup import constraint
from typer_cloup.testing import CliRunner, columns_match

runner = CliRunner()

app = typer.Typer()


@app.command()
@constraint(mutually_exclusive, "foo", "bar")
@constraint(RequireAtLeast(1), "pippo", "pluto")
def main(
    foo: Optional[str] = typer.Option(
        None, help="This text should describe the option --foo."
    ),
    bar: Optional[str] = typer.Option(
        None, help="This text should describe the option --bar."
    ),
    pippo: Optional[str] = typer.Option(
        None, help="This text should describe the option --pippo."
    ),
    pluto: Optional[str] = typer.Option(
        None, help="This text should describe the option --pluto."
    ),
):
    """This is the command description."""
    typer.echo(f"foo: {foo}")
    typer.echo(f"bar: {bar}")
    typer.echo(f"pippo: {pippo}")
    typer.echo(f"pluto: {pluto}")


def test_help():
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "[OPTIONS]" in result.output
    assert "This is the command description." in result.output
    assert "[mutually exclusive]" not in result.output
    assert columns_match(
        result.output, "--foo TEXT", "This text should describe the option --foo."
    )
    assert columns_match(
        result.output, "--bar TEXT", "This text should describe the option --bar."
    )
    assert "[at least 1 required]" not in result.output
    assert columns_match(
        result.output, "--pippo TEXT", "This text should describe the option --pippo."
    )
    assert columns_match(
        result.output, "--pluto TEXT", "This text should describe the option --pluto."
    )


def test_output():
    result = runner.invoke(
        app, ["--foo", "foo", "--pippo", "pippo", "--pluto", "pluto"]
    )
    assert result.exit_code == 0
    assert "foo: foo" in result.output
    assert "bar: None" in result.output
    assert "pippo: pippo" in result.output
    assert "pluto: pluto" in result.output


def test_too_many_options():
    result = runner.invoke(app, ["--foo", "foo", "--bar", "bar"])
    assert result.exit_code != 0
    assert "Error: the following parameters are mutually exclusive:" in result.output
    assert "--foo" in result.output
    assert "--bar" in result.output


def test_too_few_options():
    result = runner.invoke(app, [])
    assert result.exit_code != 0
    assert "Error: at least 1 of the following parameters must be set:" in result.output
    assert "--pippo" in result.output
    assert "--pluto" in result.output



import typer_cloup as typer


def main(config: typer.FileText = typer.Option(...)):
    for line in config:
        typer.echo(f"Config line: {line}")


if __name__ == "__main__":
    typer.run(main)



import typer_cloup as typer
from typer_cloup.testing import CliRunner, columns_match

runner = CliRunner()

app = typer.Typer()


@app.command()
def cmd(
    arg1: float,
    arg2: str = typer.Argument("bar"),
    opt1: str = typer.Option("foo"),
    opt2: bool = typer.Option(False),
    opt3: int = typer.Option(1),
):
    """
    Some command

    :param opt1: First option
    :param opt2: Second option
    :param opt3:
        Third
        option
    :param arg1: First argument
    :param arg2: Second argument
    """


@app.callback()
def main(global_opt: bool = False):
    """
    Callback

    :param global_opt: Global option
    """


def test_help_main():
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "Callback" in result.output
    assert ":param" not in result.output
    assert columns_match(
        result.output,
        "--global-opt / --no-global-opt",
        "Global option  [default: no-global-opt]",
    )


def test_help_cmd():
    result = runner.invoke(app, ["cmd", "--help"])
    assert result.exit_code == 0
    assert "Some command" in result.output
    assert ":param" not in result.output
    assert columns_match(result.output, "ARG1", "First argument  [required]")
    assert columns_match(result.output, "[ARG2]", "Second argument  [default: bar]")
    assert columns_match(result.output, "--opt1 TEXT", "First option  [default: foo]")
    assert columns_match(
        result.output, "--opt2 / --no-opt2", "Second option  [default: no-opt2]"
    )
    assert columns_match(result.output, "--opt3 INTEGER", "Third option  [default: 1]")



import os
import subprocess

from docs_src.options_autocompletion import tutorial002 as mod
from typer_cloup.testing import CliRunner

runner = CliRunner()


def test_completion():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, " "],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TUTORIAL002.PY_COMPLETE": "zsh_complete",
            "COMP_WORDS": "tutorial002.py --name ",
            "COMP_CWORD": "2",
            "_TYPER_COMPLETE_TESTING": "True",
        },
    )
    assert (
        "plain\nCamila\n_\n"
        "plain\nCarlos\n_\n"
        "plain\nSebastian\n_\n" in result.stdout
    )


def test_1():
    result = runner.invoke(mod.app, ["--name", "Camila"])
    assert result.exit_code == 0
    assert "Hello Camila" in result.output


def test_script():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, "--help"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
    )
    assert "Usage" in result.stdout



from typing import Optional

import typer_cloup as typer


def main(name: Optional[str] = typer.Argument(None)):
    if name is None:
        typer.echo("Hello World!")
    else:
        typer.echo(f"Hello {name}")


if __name__ == "__main__":
    typer.run(main)



import click

import typer_cloup as typer
from typer_cloup.testing import CliRunner, columns_match

runner = CliRunner()

app = typer.Typer(show_subcommand_aliases=True)


@app.callback()
def callback():
    """A package installer."""


@app.command(aliases=["i", "add"])
def install(pkg: str):
    """Install a package."""
    typer.echo(f"install: {pkg}")


# Aliases works even if cls is not a Cloup command class
@app.command(aliases=["uni", "rm"], cls=click.Command)
def uninstall(pkg: str):
    """Uninstall a package."""
    typer.echo(f"uninstall: {pkg}")


def test_help():
    result = runner.invoke(app, ["--help"])
    print(result.output)
    assert result.exit_code == 0
    assert "A package installer." in result.output
    assert columns_match(result.output, "install (i, add)", "Install a package.")
    assert columns_match(result.output, "uninstall (uni, rm)", "Uninstall a package.")


def test_install():
    result = runner.invoke(app, ["install", "foo"])
    assert result.exit_code == 0
    assert "install: foo" in result.output


def test_i():
    result = runner.invoke(app, ["i", "foo"])
    assert result.exit_code == 0
    assert "install: foo" in result.output


def test_add():
    result = runner.invoke(app, ["add", "foo"])
    assert result.exit_code == 0
    assert "install: foo" in result.output


def test_uninstall():
    result = runner.invoke(app, ["uninstall", "foo"])
    assert result.exit_code == 0
    assert "uninstall: foo" in result.output


def test_uni():
    result = runner.invoke(app, ["uni", "foo"])
    assert result.exit_code == 0
    assert "uninstall: foo" in result.output


def test_rm():
    result = runner.invoke(app, ["rm", "foo"])
    assert result.exit_code == 0
    assert "uninstall: foo" in result.output


def test_group():
    app = typer.Typer(show_subcommand_aliases=True)
    foo_app = typer.Typer(name="foo", aliases=["bar"], show_subcommand_aliases=True)

    app.add_sub(foo_app)

    @foo_app.command()
    def main():
        typer.echo("Hello World")

    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "foo (bar)" in result.stdout

    result = runner.invoke(app, ["foo", "main"])
    assert result.exit_code == 0
    assert "Hello World" in result.stdout

    result = runner.invoke(app, ["bar", "main"])
    assert result.exit_code == 0
    assert "Hello World" in result.stdout



import typer_cloup as typer


def main(config: typer.FileText = typer.Option(..., mode="a")):
    config.write("This is a single line\n")
    typer.echo("Config line written")


if __name__ == "__main__":
    typer.run(main)



import subprocess

from docs_src.testing.app03 import main as mod
from docs_src.testing.app03.test_main import test_app


def test_app03():
    test_app()


def test_script():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, "--help"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
    )
    assert "Usage" in result.stdout



import inspect
from datetime import datetime
from enum import Enum
from functools import update_wrapper
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)
from uuid import UUID

import click
import cloup

try:
    import docstring_parser
except ImportError:  # pragma: no cover
    docstring_parser = None  # type: ignore

from .completion import get_completion_inspect_parameters
from .core import TyperArgument, TyperCommand, TyperGroup, TyperOption
from .models import (
    AnyType,
    ArgumentInfo,
    CommandFunctionType,
    CommandInfo,
    Default,
    DefaultPlaceholder,
    FileBinaryRead,
    FileBinaryWrite,
    FileText,
    FileTextWrite,
    NoneType,
    OptionInfo,
    ParameterInfo,
    ParamMeta,
    Required,
    TyperInfo,
)
from .param_types import SHELL_QUOTED_LIST
from .utils import get_params_from_function

if TYPE_CHECKING:  # pragma: no cover
    from cloup.constraints import BoundConstraintSpec


def get_install_completion_arguments() -> Tuple[click.Parameter, click.Parameter]:
    install_param, show_param = get_completion_inspect_parameters()
    click_install_param, _ = get_click_param(install_param)
    click_show_param, _ = get_click_param(show_param)
    return click_install_param, click_show_param


class Typer:
    """A CLI application based on type hints."""

    def __init__(
        self,
        *,
        name: Optional[str] = Default(None),
        cls: Optional[Type[TyperGroup]] = Default(None),
        invoke_without_command: bool = Default(False),
        no_args_is_help: bool = Default(False),
        subcommand_metavar: Optional[str] = Default(None),
        chain: bool = Default(False),
        result_callback: Optional[Callable[..., Any]] = Default(None),
        align_sections: Optional[bool] = Default(None),
        show_subcommand_aliases: Optional[bool] = Default(None),
        # Command
        aliases: Optional[Iterable[str]] = Default(None),
        section: Optional[cloup.Section] = Default(None),
        constraints: Sequence["BoundConstraintSpec"] = Default(()),
        context_settings: Optional[Dict[Any, Any]] = Default(None),
        formatter_settings: Optional[Dict[str, Any]] = Default(None),
        callback: Optional[Callable[..., Any]] = Default(None),
        help: Optional[str] = Default(None),
        epilog: Optional[str] = Default(None),
        short_help: Optional[str] = Default(None),
        options_metavar: str = Default("[OPTIONS]"),
        add_help_option: bool = Default(True),
        hidden: bool = Default(False),
        deprecated: bool = Default(False),
        align_option_groups: Optional[bool] = Default(None),
        show_constraints: Optional[bool] = Default(None),
        # Completion
        add_completion: bool = True,
    ):
        self._add_completion = add_completion
        self.info = TyperInfo(
            self,
            name=name,
            cls=cls,
            invoke_without_command=invoke_without_command,
            subcommand_metavar=subcommand_metavar,
            chain=chain,
            result_callback=result_callback,
            align_sections=align_sections,
            show_subcommand_aliases=show_subcommand_aliases,
            aliases=aliases,
            section=section,
            constraints=constraints,
            context_settings=context_settings,
            formatter_settings=formatter_settings,
            callback=callback,
            help=help,
            epilog=epilog,
            short_help=short_help,
            options_metavar=options_metavar,
            add_help_option=add_help_option,
            no_args_is_help=no_args_is_help,
            hidden=hidden,
            deprecated=deprecated,
            align_option_groups=align_option_groups,
            show_constraints=show_constraints,
        )
        self.registered_groups: List[TyperInfo] = []
        self.registered_commands: List[CommandInfo] = []
        self.registered_callback: Optional[TyperInfo] = None

    def callback(
        self,
        name: Optional[str] = Default(None),
        *,
        cls: Optional[Type[TyperGroup]] = Default(None),
        invoke_without_command: bool = Default(False),
        subcommand_metavar: Optional[str] = Default(None),
        chain: bool = Default(False),
        result_callback: Optional[Callable[..., Any]] = Default(None),
        # Command
        aliases: Optional[Iterable[str]] = Default(None),
        constraints: Sequence["BoundConstraintSpec"] = Default(()),
        context_settings: Optional[Dict[Any, Any]] = Default(None),
        formatter_settings: Optional[Dict[str, Any]] = Default(None),
        help: Optional[str] = Default(None),
        epilog: Optional[str] = Default(None),
        short_help: Optional[str] = Default(None),
        options_metavar: str = Default("[OPTIONS]"),
        add_help_option: bool = Default(True),
        no_args_is_help: bool = Default(False),
        hidden: bool = Default(False),
        deprecated: bool = Default(False),
        align_option_groups: Optional[bool] = Default(None),
        show_constraints: Optional[bool] = Default(None),
    ) -> Callable[[CommandFunctionType], CommandFunctionType]:
        """Callback function decorator."""

        def decorator(f: CommandFunctionType) -> CommandFunctionType:
            info = TyperInfo(
                self,
                name=name,
                cls=cls,
                invoke_without_command=invoke_without_command,
                subcommand_metavar=subcommand_metavar,
                chain=chain,
                result_callback=result_callback,
                aliases=aliases,
                constraints=constraints,
                context_settings=context_settings,
                formatter_settings=formatter_settings,
                callback=f,
                help=help,
                epilog=epilog,
                short_help=short_help,
                options_metavar=options_metavar,
                add_help_option=add_help_option,
                no_args_is_help=no_args_is_help,
                hidden=hidden,
                deprecated=deprecated,
                align_option_groups=align_option_groups,
                show_constraints=show_constraints,
            )
            self.registered_callback = info
            cast(Any, f)._typer_info = self
            return f

        return decorator

    def command(
        self,
        name: Optional[str] = None,
        *,
        cls: Optional[Type[TyperCommand]] = None,
        aliases: Optional[Iterable[str]] = None,
        section: Optional[cloup.Section] = None,
        constraints: Sequence["BoundConstraintSpec"] = (),
        context_settings: Optional[Dict[Any, Any]] = None,
        formatter_settings: Optional[Dict[str, Any]] = None,
        help: Optional[str] = None,
        epilog: Optional[str] = None,
        short_help: Optional[str] = None,
        options_metavar: str = "[OPTIONS]",
        add_help_option: bool = True,
        no_args_is_help: bool = False,
        hidden: bool = False,
        deprecated: bool = False,
        align_option_groups: Optional[bool] = None,
        show_constraints: Optional[bool] = None,
    ) -> Callable[[CommandFunctionType], CommandFunctionType]:
        """Command function decorator"""
        if cls is None:
            cls = TyperCommand

        def decorator(f: CommandFunctionType) -> CommandFunctionType:
            info = CommandInfo(
                name=name,
                cls=cls,
                aliases=aliases,
                section=section,
                constraints=constraints,
                context_settings=context_settings,
                formatter_settings=formatter_settings,
                callback=f,
                help=help,
                epilog=epilog,
                short_help=short_help,
                options_metavar=options_metavar,
                add_help_option=add_help_option,
                no_args_is_help=no_args_is_help,
                hidden=hidden,
                deprecated=deprecated,
                align_option_groups=align_option_groups,
                show_constraints=show_constraints,
            )
            self.registered_commands.append(info)
            cast(Any, f)._command_info = info
            return f

        return decorator

    def add_sub(
        self,
        typer_instance: "Typer",
        *,
        name: Optional[str] = Default(None),
        cls: Optional[Type[TyperGroup]] = Default(None),
        invoke_without_command: bool = Default(False),
        subcommand_metavar: Optional[str] = Default(None),
        chain: bool = Default(False),
        result_callback: Optional[Callable[..., Any]] = Default(None),
        # Command
        aliases: Optional[Iterable[str]] = Default(None),
        section: Optional[cloup.Section] = Default(None),
        constraints: Sequence["BoundConstraintSpec"] = Default(()),
        context_settings: Optional[Dict[Any, Any]] = Default(None),
        formatter_settings: Optional[Dict[str, Any]] = Default(None),
        callback: Optional[Callable[..., Any]] = Default(None),
        help: Optional[str] = Default(None),
        epilog: Optional[str] = Default(None),
        short_help: Optional[str] = Default(None),
        options_metavar: str = Default("[OPTIONS]"),
        add_help_option: bool = Default(True),
        no_args_is_help: bool = Default(False),
        hidden: bool = Default(False),
        deprecated: bool = Default(False),
        align_option_groups: Optional[bool] = Default(None),
        show_constraints: Optional[bool] = Default(None),
    ) -> None:
        """Add the given sub-application to this application."""
        self.registered_groups.append(
            TyperInfo(
                typer_instance,
                name=name,
                cls=cls,
                invoke_without_command=invoke_without_command,
                no_args_is_help=no_args_is_help,
                subcommand_metavar=subcommand_metavar,
                chain=chain,
                result_callback=result_callback,
                aliases=aliases,
                section=section,
                constraints=constraints,
                context_settings=context_settings,
                formatter_settings=formatter_settings,
                callback=callback,
                help=help,
                epilog=epilog,
                short_help=short_help,
                options_metavar=options_metavar,
                add_help_option=add_help_option,
                hidden=hidden,
                deprecated=deprecated,
                show_constraints=show_constraints,
                align_option_groups=align_option_groups,
            )
        )

    def invoke(
        self,
        callback: Callable[..., Any],
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """
        Invoke a callable that is a registered command or subcommand of the given `Typer` instance.

        Other arguments are passed on to `click.Context.invoke`.
        """
        info = getattr(callback, "_command_info", None)
        if not info:
            raise TypeError("callback is not a command")

        command = get_command_from_info(info)

        ctx = click.get_current_context()
        return ctx.invoke(command, *args, **kwargs)

    def forward(
        self,
        callback: Callable[..., Any],
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """
        Forward a callable that is a registered command or subcommand of the given `Typer` instance.

        Other arguments are passed on to `click.Context.forward`.
        """
        info = getattr(callback, "_command_info", None)
        if not info:
            raise TypeError("callback is not a command")

        command = get_command_from_info(info)

        ctx = click.get_current_context()
        return ctx.forward(command, *args, **kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return get_command(self)(*args, **kwargs)


def process_help_text(command: click.Command) -> None:
    if not docstring_parser:  # pragma: no cover
        return None

    if command.help is None:
        return

    docstring = docstring_parser.parse(command.help)

    if docstring.short_description:
        command.help = docstring.short_description
        if docstring.long_description:
            command.help += "\n\n" + docstring.long_description

    for doc_param in docstring.params:
        param = next(
            (param for param in command.params if param.name == doc_param.arg_name),
            None,
        )

        help = doc_param.description
        if help is not None:
            help = help.replace("\n", " ")

        if isinstance(param, click.Option):
            if not param.help:
                param.help = help
        elif isinstance(param, TyperArgument):
            if not param.help:
                param.help = help


def get_group(typer_instance: Typer) -> click.Group:
    group = get_group_from_info(TyperInfo(typer_instance))
    return group


def get_command(typer_instance: Typer) -> click.Command:
    if typer_instance._add_completion:
        click_install_param, click_show_param = get_install_completion_arguments()
    if (
        typer_instance.registered_callback
        or typer_instance.info.callback
        or typer_instance.registered_groups
        or len(typer_instance.registered_commands) > 1
    ):
        # Create a Group
        click_command = cast(click.Command, get_group(typer_instance))
        if typer_instance._add_completion:
            click_command.params.append(click_install_param)
            click_command.params.append(click_show_param)
        return click_command
    elif len(typer_instance.registered_commands) == 1:
        # Create a single Command
        single_command = typer_instance.registered_commands[0]

        if not single_command.context_settings and not isinstance(
            typer_instance.info.context_settings, DefaultPlaceholder
        ):
            single_command.context_settings = typer_instance.info.context_settings

        click_command = get_command_from_info(single_command)
        if typer_instance._add_completion:
            click_command.params.append(click_install_param)
            click_command.params.append(click_show_param)
        return click_command
    assert False, "Could not get a command for this Typer instance"  # pragma no cover


def get_group_name(typer_info: TyperInfo) -> Optional[str]:
    if typer_info.callback:
        # Priority 1: Callback passed in app.add_sub()
        return get_command_name(typer_info.callback.__name__)
    if typer_info.typer_instance:
        registered_callback = typer_info.typer_instance.registered_callback
        if registered_callback:
            if registered_callback.callback:
                # Priority 2: Callback passed in @subapp.callback()
                return get_command_name(registered_callback.callback.__name__)
        if typer_info.typer_instance.info.callback:
            return get_command_name(typer_info.typer_instance.info.callback.__name__)
    return None


def solve_typer_info_help(typer_info: TyperInfo) -> str:
    # Priority 1: Explicit value was set in app.add_sub()
    if not isinstance(typer_info.help, DefaultPlaceholder):
        return inspect.cleandoc(typer_info.help or "")
    # Priority 2: Explicit value was set in sub_app.callback()
    try:
        callback_help = typer_info.typer_instance.registered_callback.help
        if not isinstance(callback_help, DefaultPlaceholder):
            return inspect.cleandoc(callback_help or "")
    except AttributeError:
        pass
    # Priority 3: Explicit value was set in sub_app = typer.Typer()
    try:
        instance_help = typer_info.typer_instance.info.help
        if not isinstance(instance_help, DefaultPlaceholder):
            return inspect.cleandoc(instance_help or "")
    except AttributeError:
        pass
    # Priority 4: Implicit inference from callback docstring in app.add_sub()
    if typer_info.callback:
        doc = inspect.getdoc(typer_info.callback)
        if doc:
            return doc
    # Priority 5: Implicit inference from callback docstring in @app.callback()
    try:
        callback = typer_info.typer_instance.registered_callback.callback
        if not isinstance(callback, DefaultPlaceholder):
            doc = inspect.getdoc(callback or "")
            if doc:
                return doc
    except AttributeError:
        pass
    # Priority 6: Implicit inference from callback docstring in typer.Typer()
    try:
        instance_callback = typer_info.typer_instance.info.callback
        if not isinstance(instance_callback, DefaultPlaceholder):
            doc = inspect.getdoc(instance_callback)
            if doc:
                return doc
    except AttributeError:
        pass
    # Value not set, use the default
    return typer_info.help.value


def solve_typer_info_defaults(typer_info: TyperInfo) -> TyperInfo:
    values: Dict[str, Any] = {}
    name = None
    for name, value in typer_info.__dict__.items():
        # Priority 1: Value was set in app.add_sub()
        if not isinstance(value, DefaultPlaceholder):
            values[name] = value
            continue
        # Priority 2: Value was set in @subapp.callback()
        try:
            callback_value = getattr(
                typer_info.typer_instance.registered_callback, name  # type: ignore
            )
            if not isinstance(callback_value, DefaultPlaceholder):
                values[name] = callback_value
                continue
        except AttributeError:
            pass
        # Priority 3: Value set in subapp = typer.Typer()
        try:
            instance_value = getattr(typer_info.typer_instance.info, name)  # type: ignore
            if not isinstance(instance_value, DefaultPlaceholder):
                values[name] = instance_value
                continue
        except AttributeError:
            pass
        # Value not set, use the default
        values[name] = value.value
    if values["name"] is None:
        values["name"] = get_group_name(typer_info)
    values["help"] = solve_typer_info_help(typer_info)
    return TyperInfo(**values)


def get_group_from_info(group_info: TyperInfo) -> click.Group:
    assert (
        group_info.typer_instance
    ), "A Typer instance is needed to generate a Click Group"
    commands: Dict[str, click.Command] = {}
    sections = set()
    for command_info in group_info.typer_instance.registered_commands:
        command = get_command_from_info(command_info=command_info)
        if command.name:
            commands[command.name] = command
            if isinstance(command, cloup.Command) and command_info.section is not None:
                command_info.section.add_command(command, command.name)
                sections.add(command_info.section)
    for sub_group_info in group_info.typer_instance.registered_groups:
        sub_group = get_group_from_info(sub_group_info)
        if sub_group.name:
            commands[sub_group.name] = sub_group
    solved_info = solve_typer_info_defaults(group_info)
    (
        params,
        convertors,
        context_param_name,
    ) = get_params_convertors_ctx_param_name_from_function(solved_info.callback)
    cls = cast(Optional[Type[click.Group]], solved_info.cls) or TyperGroup
    kwargs: Dict[str, Any] = {}
    if issubclass(cls, cloup.Group):
        kwargs = dict(
            sections=sections,
            align_sections=solved_info.align_sections,
            show_subcommand_aliases=solved_info.show_subcommand_aliases,
            constraints=get_constraints(
                callback=solved_info.callback,
                constraints=solved_info.constraints,
            ),
            show_constraints=solved_info.show_constraints,
            align_option_groups=solved_info.align_option_groups,
            formatter_settings=solved_info.formatter_settings,
        )
    group = cls(
        name=solved_info.name or "",
        commands=commands,
        invoke_without_command=solved_info.invoke_without_command,
        no_args_is_help=solved_info.no_args_is_help,
        subcommand_metavar=solved_info.subcommand_metavar,
        chain=solved_info.chain,
        result_callback=solved_info.result_callback,
        context_settings=solved_info.context_settings,
        callback=get_callback(
            callback=solved_info.callback,
            params=params,
            context_param_name=context_param_name,
        ),
        params=params,
        help=solved_info.help,
        epilog=solved_info.epilog,
        short_help=solved_info.short_help,
        options_metavar=solved_info.options_metavar,
        add_help_option=solved_info.add_help_option,
        hidden=solved_info.hidden,
        deprecated=solved_info.deprecated,
        **kwargs,
    )
    if solved_info.aliases:
        group.aliases = list(solved_info.aliases)  # type: ignore
    for name, command in commands.items():
        aliases = getattr(command, "aliases") or ()
        for alias in aliases:
            group.alias2name[alias] = name  # type: ignore
    process_help_text(group)
    return group


def get_command_name(name: str) -> str:
    return name.lower().replace("_", "-")


def get_params_convertors_ctx_param_name_from_function(
    callback: Optional[Callable[..., Any]]
) -> Tuple[List[Union[click.Argument, click.Option]], Dict[str, Any], Optional[str]]:
    params = {}
    convertors = {}
    context_param_name = None
    if callback:
        parameters = get_params_from_function(callback)
        for param_name, param in parameters.items():
            if lenient_issubclass(param.annotation, click.Context):
                context_param_name = param_name
                continue
            click_param, convertor = get_click_param(param)
            if convertor:
                convertors[param_name] = convertor
            params[param_name] = click_param
        callback_obj = cast(Any, callback)
        if hasattr(callback_obj, "__opt_groups"):
            for (opt_group, opt_group_params) in reversed(callback_obj.__opt_groups):
                for param_name in opt_group_params:
                    option = params.get(param_name, None)
                    if option is None:
                        raise ValueError(f"`{param_name}` is not a parameter")
                    if not isinstance(option, click.Option):
                        raise TypeError(
                            "only parameter of type `Option` can be added to option groups"
                        )
                    option_obj = cast(Any, option)
                    if getattr(option_obj, "group", None) is not None:
                        raise ValueError(
                            f"`{option}` was first assigned to `{option_obj.group}` and then "
                            f"passed as argument to `@option_group({opt_group.title!r}, ...)`"
                        )
                    option_obj.group = opt_group
                    if opt_group.hidden:
                        option.hidden = True
    return list(params.values()), convertors, context_param_name


def get_command_from_info(command_info: CommandInfo) -> click.Command:
    assert command_info.callback, "A command must have a callback function"
    name = command_info.name or get_command_name(command_info.callback.__name__)
    use_help = command_info.help
    if use_help is None:
        use_help = inspect.getdoc(command_info.callback)
    else:
        use_help = inspect.cleandoc(use_help)
    (
        params,
        convertors,
        context_param_name,
    ) = get_params_convertors_ctx_param_name_from_function(command_info.callback)
    cls = command_info.cls or TyperCommand
    kwargs: Dict[str, Any] = {}
    if issubclass(cls, cloup.Command):
        kwargs = dict(
            constraints=get_constraints(
                callback=command_info.callback,
                constraints=command_info.constraints,
            ),
            show_constraints=command_info.show_constraints,
            align_option_groups=command_info.align_option_groups,
            formatter_settings=command_info.formatter_settings,
        )
    command = cls(
        name=name,
        context_settings=command_info.context_settings,
        callback=get_callback(
            callback=command_info.callback,
            params=params,
            context_param_name=context_param_name,
        ),
        params=params,
        help=use_help,
        epilog=command_info.epilog,
        short_help=command_info.short_help,
        options_metavar=command_info.options_metavar,
        add_help_option=command_info.add_help_option,
        no_args_is_help=command_info.no_args_is_help,
        hidden=command_info.hidden,
        deprecated=command_info.deprecated,
        **kwargs,
    )
    if command_info.aliases:
        command.aliases = list(command_info.aliases)
    process_help_text(command)
    return command


def determine_type_convertor(type_: Any) -> Optional[Callable[[Any], Any]]:
    convertor: Optional[Callable[[Any], Any]] = None
    if lenient_issubclass(type_, Path):
        convertor = param_path_convertor
    if lenient_issubclass(type_, Enum):
        convertor = generate_enum_convertor(type_)
    return convertor


def param_path_convertor(value: Optional[str] = None) -> Optional[Path]:
    if value is not None:
        return Path(value)
    return None


def generate_enum_convertor(enum: Type[Enum]) -> Callable[[Any], Any]:
    lower_val_map = {str(val.value).lower(): val for val in enum}

    def convertor(value: Any) -> Any:
        if value is not None:
            low = str(value).lower()
            if low in lower_val_map:
                key = lower_val_map[low]
                return enum(key)

    return convertor


def generate_list_convertor(
    convertor: Optional[Callable[[Any], Any]]
) -> Callable[[Sequence[Any]], List[Any]]:
    def internal_convertor(value: Sequence[Any]) -> List[Any]:
        return [convertor(v) if convertor else v for v in value]

    return internal_convertor


def generate_tuple_convertor(
    types: Sequence[Any],
) -> Callable[[Tuple[Any, ...]], Tuple[Any, ...]]:
    convertors = [determine_type_convertor(type_) for type_ in types]

    def internal_convertor(param_args: Tuple[Any, ...]) -> Tuple[Any, ...]:
        return tuple(
            convertor(arg) if convertor else arg
            for (convertor, arg) in zip(convertors, param_args)
        )

    return internal_convertor


def get_callback(
    *,
    callback: Optional[Callable[..., Any]] = None,
    params: Sequence[click.Parameter] = [],
    context_param_name: Optional[str] = None,
) -> Optional[Callable[..., Any]]:
    if not callback:
        return None
    parameters = get_params_from_function(callback)
    use_params: Dict[str, Any] = {}
    for param_name in parameters:
        use_params[param_name] = None
    for param in params:
        if param.name:
            use_params[param.name] = param.default

    def wrapper(**kwargs: Any) -> Any:
        for k, v in kwargs.items():
            use_params[k] = v
        if context_param_name:
            use_params[context_param_name] = click.get_current_context()
        return callback(**use_params)  # type: ignore

    update_wrapper(wrapper, callback)
    return wrapper


def get_constraints(
    *,
    callback: Optional[Callable[..., Any]] = None,
    constraints: Sequence["BoundConstraintSpec"] = (),
) -> Sequence["BoundConstraintSpec"]:
    if not callback:
        return constraints
    callback_obj = cast(Any, callback)
    if hasattr(callback_obj, "__cloup_constraints__"):
        decorator_constraints = tuple(reversed(callback_obj.__cloup_constraints__))
        return tuple(constraints) + decorator_constraints

    return constraints


def get_click_type(
    *, annotation: Any, parameter_info: ParameterInfo
) -> click.ParamType:
    if annotation == Any:
        return click.UNPROCESSED
    if annotation == str:
        return click.STRING
    elif annotation == int:
        if parameter_info.min is not None or parameter_info.max is not None:
            min_ = None
            max_ = None
            if parameter_info.min is not None:
                min_ = int(parameter_info.min)
            if parameter_info.max is not None:
                max_ = int(parameter_info.max)
            return click.IntRange(min=min_, max=max_, clamp=parameter_info.clamp)
        else:
            return click.INT
    elif annotation == float:
        if parameter_info.min is not None or parameter_info.max is not None:
            return click.FloatRange(
                min=parameter_info.min,
                max=parameter_info.max,
                clamp=parameter_info.clamp,
            )
        else:
            return click.FLOAT
    elif annotation == bool:
        return click.BOOL
    elif annotation == UUID:
        return click.UUID
    elif annotation == datetime:
        return click.DateTime(formats=parameter_info.formats)
    elif (
        annotation == Path
        or parameter_info.allow_dash
        or parameter_info.path_type
        or parameter_info.resolve_path
    ):
        return click.Path(
            exists=parameter_info.exists,
            file_okay=parameter_info.file_okay,
            dir_okay=parameter_info.dir_okay,
            writable=parameter_info.writable,
            readable=parameter_info.readable,
            resolve_path=parameter_info.resolve_path,
            allow_dash=parameter_info.allow_dash,
            path_type=parameter_info.path_type,
        )
    elif lenient_issubclass(annotation, FileTextWrite):
        return click.File(
            mode=parameter_info.mode or "w",
            encoding=parameter_info.encoding,
            errors=parameter_info.errors,
            lazy=parameter_info.lazy,
            atomic=parameter_info.atomic,
        )
    elif lenient_issubclass(annotation, FileText):
        return click.File(
            mode=parameter_info.mode or "r",
            encoding=parameter_info.encoding,
            errors=parameter_info.errors,
            lazy=parameter_info.lazy,
            atomic=parameter_info.atomic,
        )
    elif lenient_issubclass(annotation, FileBinaryRead):
        return click.File(
            mode=parameter_info.mode or "rb",
            encoding=parameter_info.encoding,
            errors=parameter_info.errors,
            lazy=parameter_info.lazy,
            atomic=parameter_info.atomic,
        )
    elif lenient_issubclass(annotation, FileBinaryWrite):
        return click.File(
            mode=parameter_info.mode or "wb",
            encoding=parameter_info.encoding,
            errors=parameter_info.errors,
            lazy=parameter_info.lazy,
            atomic=parameter_info.atomic,
        )
    elif lenient_issubclass(annotation, Enum):
        return click.Choice(
            [item.value for item in annotation],
            case_sensitive=parameter_info.case_sensitive,
        )
    raise RuntimeError(f"Type not yet supported: {annotation}")  # pragma no cover


def lenient_issubclass(
    cls: Any, class_or_tuple: Union[AnyType, Tuple[AnyType, ...]]
) -> bool:
    return isinstance(cls, type) and issubclass(cls, class_or_tuple)


def get_click_param(
    param: ParamMeta,
) -> Tuple[Union[click.Argument, click.Option], Any]:
    # First, find out what will be:
    # * ParamInfo (ArgumentInfo or OptionInfo)
    # * default_value
    # * required
    default_value = None
    required = False
    if isinstance(param.default, ParameterInfo):
        parameter_info = param.default
        if parameter_info.default == Required:
            required = True
        else:
            default_value = parameter_info.default
    elif param.default == Required or param.default == param.empty:
        required = True
        parameter_info = ArgumentInfo()
    else:
        default_value = param.default
        parameter_info = OptionInfo()
    annotation: Any = Any
    if not param.annotation == param.empty:
        annotation = param.annotation
    else:
        annotation = str
    main_type = annotation
    is_list = False
    is_tuple = False
    parameter_type: Any = None
    is_flag = None
    origin = getattr(main_type, "__origin__", None)
    if origin is not None:
        # Handle Optional[SomeType]
        if origin is Union:
            types = []
            for type_ in main_type.__args__:
                if type_ is NoneType:
                    continue
                types.append(type_)
            assert len(types) == 1, "Typer currently doesn't support `Union` types"
            main_type = types[0]
            origin = getattr(main_type, "__origin__", None)
        # Handle Tuples and Lists
        if lenient_issubclass(origin, List):
            main_type = main_type.__args__[0]
            assert not getattr(
                main_type, "__origin__", None
            ), "List types with complex sub-types are not currently supported"
            if parameter_info.shell_quoted_list:
                parameter_type = SHELL_QUOTED_LIST
            else:
                is_list = True
        elif lenient_issubclass(origin, Tuple):  # type: ignore
            types = []
            for type_ in main_type.__args__:
                assert not getattr(
                    type_, "__origin__", None
                ), "Tuple types with complex sub-types are not currently supported"
                types.append(
                    get_click_type(annotation=type_, parameter_info=parameter_info)
                )
            parameter_type = tuple(types)
            is_tuple = True
    if parameter_type is None:
        parameter_type = get_click_type(
            annotation=main_type, parameter_info=parameter_info
        )
    convertor = determine_type_convertor(main_type)
    if is_list or parameter_type is SHELL_QUOTED_LIST:
        convertor = generate_list_convertor(convertor)
    if is_tuple:
        convertor = generate_tuple_convertor(main_type.__args__)
    if isinstance(parameter_info, OptionInfo):
        if main_type is bool and not (parameter_info.is_flag is False):
            is_flag = True
            # Click doesn't accept a flag of type bool, only None, and then it sets it
            # to bool internally
            parameter_type = None
        default_option_name = get_command_name(param.name)
        if is_flag:
            default_option_declaration = (
                f"--{default_option_name}/--no-{default_option_name}"
            )
        else:
            default_option_declaration = f"--{default_option_name}"
        param_decls = [param.name]
        if parameter_info.param_decls:
            param_decls.extend(parameter_info.param_decls)
        else:
            param_decls.append(default_option_declaration)
        return (
            TyperOption(
                # Option
                param_decls=param_decls,
                show_default=parameter_info.show_default,
                prompt=parameter_info.prompt,
                confirmation_prompt=parameter_info.confirmation_prompt,
                prompt_required=parameter_info.prompt_required,
                hide_input=parameter_info.hide_input,
                is_flag=is_flag,
                flag_value=parameter_info.flag_value,
                multiple=is_list,
                count=parameter_info.count,
                allow_from_autoenv=parameter_info.allow_from_autoenv,
                type=parameter_type,
                convertor=convertor,
                help=parameter_info.help,
                hidden=parameter_info.hidden,
                show_choices=parameter_info.show_choices,
                show_envvar=parameter_info.show_envvar,
                # TyperOption
                group=parameter_info.group,
                # Parameter
                required=required,
                default=default_value,
                callback=get_param_callback(
                    callback=parameter_info.callback, convertor=convertor
                ),
                metavar=parameter_info.metavar,
                expose_value=parameter_info.expose_value,
                is_eager=parameter_info.is_eager,
                envvar=parameter_info.envvar,
                shell_complete=get_param_completion(parameter_info.shell_complete),
            ),
            convertor,
        )
    elif isinstance(parameter_info, ArgumentInfo):
        param_decls = [param.name]
        nargs = None
        if is_list:
            nargs = -1
        return (
            TyperArgument(
                # Argument
                param_decls=param_decls,
                type=parameter_type,
                convertor=convertor,
                required=required,
                nargs=nargs,
                # TyperArgument
                show_default=parameter_info.show_default,
                show_choices=parameter_info.show_choices,
                show_envvar=parameter_info.show_envvar,
                allow_from_autoenv=parameter_info.allow_from_autoenv,
                help=parameter_info.help,
                hidden=parameter_info.hidden,
                # Parameter
                default=default_value,
                callback=get_param_callback(
                    callback=parameter_info.callback, convertor=convertor
                ),
                metavar=parameter_info.metavar,
                expose_value=parameter_info.expose_value,
                is_eager=parameter_info.is_eager,
                envvar=parameter_info.envvar,
                shell_complete=get_param_completion(parameter_info.shell_complete),
            ),
            convertor,
        )
    assert False, "A click.Parameter should be returned"  # pragma no cover


def get_param_callback(
    *,
    callback: Optional[Callable[..., Any]] = None,
    convertor: Optional[Callable[..., Any]] = None,
) -> Optional[Callable[..., Any]]:
    if not callback:
        return None
    parameters = get_params_from_function(callback)
    ctx_name = None
    click_param_name = None
    value_name = None
    untyped_names: List[str] = []
    for param_name, param_sig in parameters.items():
        if lenient_issubclass(param_sig.annotation, click.Context):
            ctx_name = param_name
        elif lenient_issubclass(param_sig.annotation, click.Parameter):
            click_param_name = param_name
        else:
            untyped_names.append(param_name)
    # Extract value param name first
    if untyped_names:
        value_name = untyped_names.pop()
    # If context and Click param were not typed (old/Click callback style) extract them
    if untyped_names:
        if ctx_name is None:
            ctx_name = untyped_names.pop(0)
        if click_param_name is None:
            if untyped_names:
                click_param_name = untyped_names.pop(0)
        if untyped_names:
            raise click.ClickException(
                "Too many CLI parameter callback function parameters"
            )

    def wrapper(ctx: click.Context, param: click.Parameter, value: Any) -> Any:
        use_params: Dict[str, Any] = {}
        if ctx_name:
            use_params[ctx_name] = ctx
        if click_param_name:
            use_params[click_param_name] = param
        if value_name:
            if convertor:
                use_value = convertor(value)
            else:
                use_value = value
            use_params[value_name] = use_value
        return callback(**use_params)  # type: ignore

    update_wrapper(wrapper, callback)
    return wrapper


def get_param_completion(
    callback: Optional[Callable[..., Any]] = None
) -> Optional[Callable[..., Any]]:
    if not callback:
        return None
    parameters = get_params_from_function(callback)
    ctx_name = None
    param_name = None
    incomplete_name = None
    unassigned_params = [param for param in parameters.values()]
    for param_sig in unassigned_params[:]:
        if lenient_issubclass(param_sig.annotation, click.Context):
            ctx_name = param_sig.name
            unassigned_params.remove(param_sig)
        elif lenient_issubclass(param_sig.annotation, click.Parameter):
            param_name = param_sig.name
            unassigned_params.remove(param_sig)
        elif lenient_issubclass(param_sig.annotation, str):
            incomplete_name = param_sig.name
            unassigned_params.remove(param_sig)
    # If there are still unassigned parameters (not typed), extract by name
    for param_sig in unassigned_params[:]:
        if ctx_name is None and param_sig.name == "ctx":
            ctx_name = param_sig.name
            unassigned_params.remove(param_sig)
        elif param_name is None and param_sig.name == "param":
            param_name = param_sig.name
            unassigned_params.remove(param_sig)
        elif incomplete_name is None and param_sig.name == "incomplete":
            incomplete_name = param_sig.name
            unassigned_params.remove(param_sig)
    # Extract value param name first
    if unassigned_params:
        show_params = " ".join([param.name for param in unassigned_params])
        raise click.ClickException(
            f"Invalid shell-completion callback parameters: {show_params}"
        )

    def wrapper(
        ctx: click.Context, param: click.Parameter, incomplete: Optional[str]
    ) -> Any:
        use_params: Dict[str, Any] = {}
        if ctx_name:
            use_params[ctx_name] = ctx
        if param_name:
            use_params[param_name] = param
        if incomplete_name:
            use_params[incomplete_name] = incomplete
        return callback(**use_params)  # type: ignore

    update_wrapper(wrapper, callback)
    return wrapper


def run(function: Callable[..., Any]) -> None:
    app = Typer(add_completion=False)
    app.command()(function)
    app()



import typer_cloup as typer

app = typer.Typer()


@app.command("create")
def cli_create_user(username: str):
    typer.echo(f"Creating user: {username}")


@app.command("delete")
def cli_delete_user(username: str):
    typer.echo(f"Deleting user: {username}")


if __name__ == "__main__":
    app()



import inspect
import os
import sys
from gettext import gettext as _
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import click
import click.core
import click.formatting
import click.parser
import click.types
import cloup

from .models import Context

if TYPE_CHECKING:  # pragma: no cover
    import click.shell_completion


class TyperParameterMixin(click.Parameter):
    convertor: Optional[Any] = None

    def __init__(self, convertor: Optional[Any] = None, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.convertor = convertor

    def handle_parse_result(
        self, ctx: click.Context, opts: Mapping[str, Any], args: List[str]
    ) -> Tuple[Any, List[str]]:
        value, args = super().handle_parse_result(ctx, opts, args)

        if self.convertor and self.expose_value:
            ctx.params[self.name] = self.convertor(ctx.params[self.name])  # type: ignore

        return value, args


class TyperArgument(TyperParameterMixin, cloup.Argument):
    """A Typer argument."""

    def __init__(
        self,
        *,
        # click.Parameter
        param_decls: List[str],
        type: Optional[Any] = None,
        required: Optional[bool] = None,
        default: Optional[Any] = None,
        callback: Optional[Callable[..., Any]] = None,
        nargs: Optional[int] = None,
        metavar: Optional[str] = None,
        expose_value: bool = True,
        is_eager: bool = False,
        envvar: Optional[Union[str, List[str]]] = None,
        shell_complete: Optional[
            Callable[
                [click.Context, click.Parameter, str],
                Union[List["click.shell_completion.CompletionItem"], List[str]],
            ]
        ] = None,
        # TyperParameterMixin
        convertor: Optional[Any] = None,
        # TyperArgument
        show_default: Union[bool, str] = True,
        show_choices: bool = True,
        show_envvar: bool = True,
        allow_from_autoenv: bool = False,
        help: Optional[str] = None,
        hidden: bool = False,
    ):
        super().__init__(
            param_decls=param_decls,
            type=type,
            required=required,
            default=default,
            callback=callback,
            nargs=nargs,
            metavar=metavar,
            expose_value=expose_value,
            is_eager=is_eager,
            envvar=envvar,
            shell_complete=shell_complete,
            convertor=convertor,
        )
        self.allow_from_autoenv = allow_from_autoenv
        self.help = help
        self.show_default = show_default
        self.show_choices = show_choices
        self.show_envvar = show_envvar
        self.hidden = hidden

    def resolve_envvar_value(self, ctx: click.Context) -> Optional[str]:
        rv = super().resolve_envvar_value(ctx)

        if rv is not None:
            return rv

        if (
            self.allow_from_autoenv
            and ctx.auto_envvar_prefix is not None
            and self.name is not None
        ):
            envvar = self.name.upper()
            if ctx.auto_envvar_prefix:
                envvar = f"{ctx.auto_envvar_prefix}_{envvar}"

            rv = os.environ.get(envvar)

            if rv:
                return rv

        return None

    def get_help_record(self, ctx: click.Context) -> Tuple[str, str]:
        # Modified version of click.core.Option.get_help_record()
        # to support arguments

        name = self.make_metavar()
        help = self.help or ""
        extra = []

        if self.show_envvar:
            envvar = self.envvar

            if envvar is None:
                if (
                    self.allow_from_autoenv
                    and ctx.auto_envvar_prefix is not None
                    and self.name is not None
                ):
                    envvar = self.name.upper()
                    if ctx.auto_envvar_prefix:
                        envvar = f"{ctx.auto_envvar_prefix}_{envvar}"

            if envvar is not None:
                var_str = (
                    ", ".join(str(d) for d in envvar)
                    if isinstance(envvar, (list, tuple))
                    else envvar
                )
                extra.append(f"env var: {var_str}")

        # Temporarily enable resilient parsing to avoid type casting
        # failing for the default. Might be possible to extend this to
        # help formatting in general.
        resilient = ctx.resilient_parsing
        ctx.resilient_parsing = True

        try:
            default_value = self.get_default(ctx, call=False)
        finally:
            ctx.resilient_parsing = resilient

        show_default_is_str = isinstance(self.show_default, str)

        if show_default_is_str or (
            default_value is not None and (self.show_default or ctx.show_default)
        ):
            if show_default_is_str:
                default_string = f"({self.show_default})"
            elif isinstance(default_value, (list, tuple)):
                default_string = ", ".join(str(d) for d in default_value)
            elif callable(default_value):
                default_string = _("(dynamic)")
            else:
                default_string = str(default_value)

            if default_string:
                extra.append(_("default: {default}").format(default=default_string))

        if isinstance(self.type, click.types._NumberRangeBase):
            range_str = self.type._describe_range()

            if range_str:
                extra.append(range_str)

        if self.required:
            extra.append(_("required"))

        if extra:
            extra_str = "; ".join(extra)
            help = f"{help}  [{extra_str}]" if help else f"[{extra_str}]"

        return name, help

    def make_metavar(self) -> str:
        # Modified version of click.core.Argument.make_metavar()
        # to include Argument name
        if self.metavar is not None:
            return self.metavar
        var = (self.name or "").upper()
        if not self.required:
            var = "[{}]".format(var)
        type_var = self.type.get_metavar(self)
        if type_var:
            var += f":{type_var}"
        if self.nargs != 1:
            var += "..."
        return var

    def shell_complete(
        self, ctx: click.Context, incomplete: str
    ) -> List["click.shell_completion.CompletionItem"]:
        return super().shell_complete(ctx, incomplete)


class TyperOption(TyperParameterMixin, cloup.Option):
    """A Typer option."""

    def __init__(
        self,
        # click.Parameter
        param_decls: List[str],
        type: Optional[Union[click.types.ParamType, Any]] = None,
        required: Optional[bool] = None,
        default: Optional[Any] = None,
        callback: Optional[Callable[..., Any]] = None,
        nargs: Optional[int] = None,
        metavar: Optional[str] = None,
        expose_value: bool = True,
        is_eager: bool = False,
        envvar: Optional[Union[str, List[str]]] = None,
        shell_complete: Optional[
            Callable[
                [click.Context, click.Parameter, str],
                Union[List["click.shell_completion.CompletionItem"], List[str]],
            ]
        ] = None,
        # TyperParameterMixin
        convertor: Optional[Any] = None,
        # click.Option
        show_default: Union[bool, str] = False,
        prompt: Union[bool, str] = False,
        confirmation_prompt: Union[bool, str] = False,
        prompt_required: bool = True,
        hide_input: bool = False,
        is_flag: Optional[bool] = None,
        flag_value: Optional[Any] = None,
        multiple: bool = False,
        count: bool = False,
        allow_from_autoenv: bool = True,
        help: Optional[str] = None,
        hidden: bool = False,
        show_choices: bool = True,
        show_envvar: bool = False,
        # cloup.Option
        group: Optional[cloup.OptionGroup] = None,
    ):
        super().__init__(
            param_decls=param_decls,
            type=type,
            required=required,
            default=default,
            callback=callback,
            nargs=nargs,
            metavar=metavar,
            expose_value=expose_value,
            is_eager=is_eager,
            envvar=envvar,
            shell_complete=shell_complete,
            convertor=convertor,
            show_default=show_default,
            prompt=prompt,
            confirmation_prompt=confirmation_prompt,
            prompt_required=prompt_required,
            hide_input=hide_input,
            is_flag=is_flag,
            flag_value=flag_value,
            multiple=multiple,
            count=count,
            allow_from_autoenv=allow_from_autoenv,
            help=help,
            hidden=hidden,
            show_choices=show_choices,
            show_envvar=show_envvar,
            group=group,
        )

    def resolve_envvar_value(self, ctx: click.Context) -> Optional[str]:
        rv = super().resolve_envvar_value(ctx)

        if rv is not None:
            return rv

        if (
            self.allow_from_autoenv
            and ctx.auto_envvar_prefix is not None
            and self.name is not None
        ):
            envvar = self.name.upper()
            if ctx.auto_envvar_prefix:
                envvar = f"{ctx.auto_envvar_prefix}_{envvar}"

            rv = os.environ.get(envvar)

            if rv:
                return rv

        return None

    def get_help_record(self, ctx: click.Context) -> Optional[Tuple[str, str]]:
        # Duplicate all of Click's logic with two exceptions:
        # 1.  handle case when `ctx.auto_envvar_prefix` is empty string
        # 2.  to allow boolean flags with only names for False values as it's currently supported by Typer
        #     Ref: https://typer-cloup.netlify.app/tutorial/parameter-types/bool/#only-names-for-false

        any_prefix_is_slash = False

        def _write_opts(opts: Sequence[str]) -> str:
            nonlocal any_prefix_is_slash

            rv, any_slashes = click.formatting.join_options(opts)

            if any_slashes:
                any_prefix_is_slash = True

            if not self.is_flag and not self.count:
                rv += f" {self.make_metavar()}"

            return rv

        rv = [_write_opts(self.opts)]

        if self.secondary_opts:
            rv.append(_write_opts(self.secondary_opts))

        help = self.help or ""
        extra = []

        if self.show_envvar:
            envvar = self.envvar

            if envvar is None:
                if (
                    self.allow_from_autoenv
                    and ctx.auto_envvar_prefix is not None
                    and self.name is not None
                ):
                    # Typer override, original commented
                    envvar = self.name.upper()
                    if ctx.auto_envvar_prefix:
                        envvar = f"{ctx.auto_envvar_prefix}_{envvar}"
                    # Typer override end

            if envvar is not None:
                var_str = (
                    envvar
                    if isinstance(envvar, str)
                    else ", ".join(str(d) for d in envvar)
                )
                extra.append(_("env var: {var}").format(var=var_str))

        # Temporarily enable resilient parsing to avoid type casting
        # failing for the default. Might be possible to extend this to
        # help formatting in general.
        resilient = ctx.resilient_parsing
        ctx.resilient_parsing = True

        try:
            default_value = self.get_default(ctx, call=False)
        finally:
            ctx.resilient_parsing = resilient

        show_default = False
        show_default_is_str = False

        if self.show_default is not None:
            if isinstance(self.show_default, str):
                show_default_is_str = show_default = True
            else:
                show_default = self.show_default
        elif ctx.show_default is not None:  # pragma: no cover
            show_default = ctx.show_default

        if show_default_is_str or (show_default and (default_value is not None)):
            if show_default_is_str:
                default_string = f"({self.show_default})"
            elif isinstance(default_value, (list, tuple)):
                default_string = ", ".join(str(d) for d in default_value)
            elif inspect.isfunction(default_value):
                default_string = _("(dynamic)")
            elif self.is_bool_flag and self.secondary_opts:
                # For boolean flags that have distinct True/False opts,
                # use the opt without prefix instead of the value.
                # Typer override, original commented
                # default_string = click.parser.split_opt(
                #     (self.opts if self.default else self.secondary_opts)[0]
                # )[1]
                if self.default:
                    if self.opts:
                        default_string = click.parser.split_opt(self.opts[0])[1]
                    else:
                        default_string = str(default_value)
                else:
                    default_string = click.parser.split_opt(self.secondary_opts[0])[1]
                # Typer override end
            elif self.is_bool_flag and not self.secondary_opts and not default_value:
                default_string = ""
            else:
                default_string = str(default_value)

            if default_string:
                extra.append(_("default: {default}").format(default=default_string))

        if (
            isinstance(self.type, click.types._NumberRangeBase)
            # skip count with default range type
            and not (self.count and self.type.min == 0 and self.type.max is None)
        ):
            range_str = self.type._describe_range()

            if range_str:
                extra.append(range_str)

        if self.required:
            extra.append(_("required"))

        if extra:
            extra_str = "; ".join(extra)
            help = f"{help}  [{extra_str}]" if help else f"[{extra_str}]"

        return ("; " if any_prefix_is_slash else " / ").join(rv), help

    def shell_complete(
        self, ctx: click.Context, incomplete: str
    ) -> List["click.shell_completion.CompletionItem"]:
        return super().shell_complete(ctx, incomplete)


def _typer_main_shell_completion(
    self: click.core.Command,
    *,
    ctx_args: Dict[str, Any],
    prog_name: str,
    complete_var: Optional[str] = None,
) -> None:
    if complete_var is None:
        complete_var = "_{}_COMPLETE".format(prog_name.replace("-", "_").upper())

    instruction = os.environ.get(complete_var)

    if not instruction:
        return

    from .completion import shell_complete

    rv = shell_complete(self, ctx_args, prog_name, complete_var, instruction)
    sys.exit(rv)


class TyperCommand(cloup.Command):
    """A Typer command."""

    context_class: Type[Context] = Context

    def _main_shell_completion(
        self,
        ctx_args: Dict[str, Any],
        prog_name: str,
        complete_var: Optional[str] = None,
    ) -> None:
        _typer_main_shell_completion(
            self, ctx_args=ctx_args, prog_name=prog_name, complete_var=complete_var
        )

    def get_arguments_help_section(
        self, ctx: click.Context
    ) -> Optional[cloup.HelpSection]:
        def is_arg_hidden(arg: click.Argument) -> bool:
            if isinstance(arg, TyperArgument):
                return arg.hidden
            return False  # pragma: no cover

        if all(is_arg_hidden(arg) for arg in self.arguments):
            return None
        return cloup.HelpSection(
            heading="Arguments",
            definitions=[
                self.get_argument_help_record(arg, ctx)
                for arg in self.arguments
                if not is_arg_hidden(arg)
            ],
        )

    def get_default_option_group(
        self, ctx: click.Context, is_the_only_visible_option_group: bool = False
    ) -> cloup.OptionGroup:
        """
        Return an `OptionGroup` instance for the options not explicitly
        assigned to an option group, eventually including the `--help` option.
        """
        default_group = cloup.OptionGroup("Options")
        default_group.options = self.get_ungrouped_options(ctx)
        return default_group


class TyperGroup(TyperCommand, cloup.Group):
    """A Typer group."""

    context_class: Type[Context] = Context



import typer_cloup as typer


def main(name: str = typer.Argument(..., help="The name of the user to greet")):
    """
    Say hi to NAME very gently, like Dirk.
    """
    typer.echo(f"Hello {name}")


if __name__ == "__main__":
    typer.run(main)



from click.shell_completion import CompletionItem

import typer_cloup as typer

app = typer.Typer()


def complete(ctx, param, incomplete):
    typer.echo(f"info name is: {ctx.info_name}", err=True)
    typer.echo(f"param is: {param.name}", err=True)
    typer.echo(f"incomplete is: {incomplete}", err=True)
    return [
        CompletionItem("Camila", help="The reader of books."),
        CompletionItem("Carlos", help="The writer of scripts."),
        CompletionItem("Sebastian", help="The type hints guy."),
    ]


@app.command()
def main(name: str = typer.Option("World", shell_complete=complete)):
    typer.echo(f"Hello {name}")


if __name__ == "__main__":
    app()



import os
import subprocess
from pathlib import Path

import typer_cloup as typer
from docs_src.commands.index import tutorial001 as mod
from typer_cloup.testing import CliRunner

runner = CliRunner()
app = typer.Typer()
app.command()(mod.main)


def test_completion_install_no_shell():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, "--install-completion"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TYPER_COMPLETE_TESTING": "True",
            "_TYPER_COMPLETE_TEST_DISABLE_SHELL_DETECTION": "True",
        },
    )
    assert "Error: Option '--install-completion' requires an argument" in result.stderr


def test_completion_install_bash():
    bash_completion_path: Path = Path.home() / ".bashrc"
    text = ""
    if bash_completion_path.is_file():
        text = bash_completion_path.read_text()
    result = subprocess.run(
        ["coverage", "run", mod.__file__, "--install-completion", "bash"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TYPER_COMPLETE_TESTING": "True",
            "_TYPER_COMPLETE_TEST_DISABLE_SHELL_DETECTION": "True",
        },
    )
    new_text = bash_completion_path.read_text()
    bash_completion_path.write_text(text)
    install_source = ".bash_completions/tutorial001.py.sh"
    assert install_source not in text
    assert install_source in new_text
    assert "completion installed at" in result.stdout
    assert "Completion will take effect once you restart the terminal" in result.stdout
    install_source_path = Path.home() / install_source
    assert install_source_path.is_file()
    install_content = install_source_path.read_text()
    install_source_path.unlink()
    assert (
        "complete -o nosort -F _tutorial001py_completion tutorial001.py"
        in install_content
    )


def test_completion_install_zsh():
    completion_path: Path = Path.home() / ".zshrc"
    text = ""
    if not completion_path.is_file():  # pragma: no cover
        completion_path.write_text('echo "custom .zshrc"')
    if completion_path.is_file():
        text = completion_path.read_text()
    result = subprocess.run(
        ["coverage", "run", mod.__file__, "--install-completion", "zsh"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TYPER_COMPLETE_TESTING": "True",
            "_TYPER_COMPLETE_TEST_DISABLE_SHELL_DETECTION": "True",
        },
    )
    new_text = completion_path.read_text()
    completion_path.write_text(text)
    zfunc_fragment = "fpath+=~/.zfunc"
    assert zfunc_fragment in new_text
    assert "completion installed at" in result.stdout
    assert "Completion will take effect once you restart the terminal" in result.stdout
    install_source_path = Path.home() / ".zfunc/_tutorial001.py"
    assert install_source_path.is_file()
    install_content = install_source_path.read_text()
    install_source_path.unlink()
    assert "compdef _tutorial001py_completion tutorial001.py" in install_content


def test_completion_install_fish():
    script_path = Path(mod.__file__)
    completion_path: Path = (
        Path.home() / f".config/fish/completions/{script_path.name}.fish"
    )
    result = subprocess.run(
        ["coverage", "run", mod.__file__, "--install-completion", "fish"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TYPER_COMPLETE_TESTING": "True",
            "_TYPER_COMPLETE_TEST_DISABLE_SHELL_DETECTION": "True",
        },
    )
    new_text = completion_path.read_text()
    completion_path.unlink()
    assert "complete --no-files --command tutorial001.py" in new_text
    assert "completion installed at" in result.stdout
    assert "Completion will take effect once you restart the terminal" in result.stdout


runner = CliRunner()
app = typer.Typer()
app.command()(mod.main)



import os
import subprocess

from docs_src.commands.index import tutorial001 as mod


def test_completion_show_no_shell():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, "--show-completion"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TYPER_COMPLETE_TESTING": "True",
            "_TYPER_COMPLETE_TEST_DISABLE_SHELL_DETECTION": "True",
        },
    )
    assert "Error: Option '--show-completion' requires an argument" in result.stderr


def test_completion_show_bash():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, "--show-completion", "bash"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TYPER_COMPLETE_TESTING": "True",
            "_TYPER_COMPLETE_TEST_DISABLE_SHELL_DETECTION": "True",
        },
    )
    assert (
        "complete -o nosort -F _tutorial001py_completion tutorial001.py"
        in result.stdout
    )


def test_completion_source_zsh():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, "--show-completion", "zsh"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TYPER_COMPLETE_TESTING": "True",
            "_TYPER_COMPLETE_TEST_DISABLE_SHELL_DETECTION": "True",
        },
    )
    assert "compdef _tutorial001py_completion tutorial001.py" in result.stdout


def test_completion_source_fish():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, "--show-completion", "fish"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TYPER_COMPLETE_TESTING": "True",
            "_TYPER_COMPLETE_TEST_DISABLE_SHELL_DETECTION": "True",
        },
    )
    assert "complete --no-files --command tutorial001.py" in result.stdout



from cloup import Section

import typer_cloup as typer
from typer_cloup.testing import CliRunner, columns_match

runner = CliRunner()

app = typer.Typer(name="git", align_sections=True)


def f(**kwargs):
    """Dummy command callback"""
    pass  # pragma: no cover


section1 = Section("Start a working area (see also: git help tutorial)")
section2 = Section("Work on the current change (see also: git help everyday)")
section3 = Section("Examine the history and state (see also: git help revisions)")


@app.command(section=section1, help="Clone a repository into a new directory")
def clone():
    pass  # pragma: no cover


@app.command(
    section=section1,
    help="Create an empty Git repository or reinitialize an existing one",
)
def init():
    pass  # pragma: no cover


@app.command(
    section=section2, help="Remove files from the working tree and from the index"
)
def rm():
    pass  # pragma: no cover


@app.command(section=section2, help="Initialize and modify the sparse-checkout")
def sparse_checkout():
    pass  # pragma: no cover


@app.command(section=section2, help="Move or rename a file, a directory, or a symlink")
def mv():
    pass  # pragma: no cover


@app.command(section=section3, help="Show the working tree status")
def status():
    pass  # pragma: no cover


@app.command(
    section=section3, help="Show changes between commits, commit and working tree, etc"
)
def diff():
    pass  # pragma: no cover


@app.command(
    section=section3, help="Use binary search to find the commit that introduced a bug"
)
def bisect():
    pass  # pragma: no cover


# The following commands will be added to the "default section" (a sorted Section)


@app.command(help="Fake command #2")
def fake_2():
    pass  # pragma: no cover


@app.command(help="Fake command #1")
def fake_1():
    pass  # pragma: no cover


def test_help():
    result = runner.invoke(app, ["--help"])
    print(result.output)
    assert result.exit_code == 0
    assert "Start a working area (see also: git help tutorial):" in result.output
    assert columns_match(
        result.output, "clone", "Clone a repository into a new directory"
    )
    assert columns_match(result.output, "init", "Create an empty Git repository")
    assert "Work on the current change (see also: git help everyday):" in result.output
    assert columns_match(
        result.output, "rm", "Remove files from the working tree and from the index"
    )
    assert columns_match(
        result.output, "sparse-checkout", "Initialize and modify the sparse-checkout"
    )
    assert columns_match(
        result.output, "mv", "Move or rename a file, a directory, or a symlink"
    )
    assert (
        "Examine the history and state (see also: git help revisions):" in result.output
    )
    assert columns_match(result.output, "status", "Show the working tree status")
    assert columns_match(result.output, "diff", "Show changes between commits")
    assert columns_match(
        result.output,
        "bisect",
        "Use binary search to find the commit that introduced a bug",
    )



import typer_cloup as typer
from typer_cloup.testing import CliRunner

runner = CliRunner()

app = typer.Typer()


@app.command()
def test(count: int = typer.Option(1)):
    typer.echo(f"Count: {count}")


@app.command()
def dist(count: int = typer.Option(1)):
    app.forward(test)
    app.invoke(test, count=42)


def test_invocation():
    result = runner.invoke(app, ["dist", "--count", "2"])
    assert result.exit_code == 0
    assert "Count: 2\nCount: 42" in result.output



import typer_cloup as typer
from typer_cloup.testing import CliRunner

runner = CliRunner()


def test_autoenv_prefix():
    app = typer.Typer(context_settings=dict(auto_envvar_prefix="TEST"))

    @app.command()
    def main(
        name: str = typer.Argument(..., allow_from_autoenv=True),
        lastname: str = typer.Option(...),
    ):
        typer.echo(f"Hello {name} {lastname}")

    result = runner.invoke(app, ["--help"])
    assert "[env var: TEST_NAME; required]" in result.stdout
    assert "[env var: TEST_LASTNAME; required]" in result.stdout

    result = runner.invoke(app, env={"TEST_NAME": "Joe", "TEST_LASTNAME": "Bloggs"})
    assert result.exit_code == 0
    assert "Hello Joe Bloggs" in result.output


def test_autoenv_prefix_empty():
    app = typer.Typer(context_settings=dict(auto_envvar_prefix=""))

    @app.command()
    def main(
        name: str = typer.Argument(..., allow_from_autoenv=True),
        lastname: str = typer.Option(...),
    ):
        typer.echo(f"Hello {name} {lastname}")

    result = runner.invoke(app, ["--help"])
    assert "[env var: NAME; required]" in result.stdout
    assert "[env var: LASTNAME; required]" in result.stdout

    result = runner.invoke(app, env={"NAME": "Joe", "LASTNAME": "Bloggs"})
    assert result.exit_code == 0
    assert "Hello Joe Bloggs" in result.output



import os
import sys
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

import click
import click.exceptions
import click.shell_completion
from click.shell_completion import (
    BashComplete,
    FishComplete,
    ZshComplete,
    get_completion_class,
)

from .models import ParamMeta
from .params import Option
from .utils import get_params_from_function

try:
    import shellingham
except ImportError:  # pragma: no cover
    shellingham = None


Shells = Enum(  # type: ignore[misc]
    "Shells",
    names={shell: shell for shell in click.shell_completion._available_shells.keys()},
    type=str,
)


def get_completion_script(
    cli: click.BaseCommand,
    ctx_args: Dict[str, Any],
    prog_name: str,
    complete_var: str,
    shell: str,
) -> str:
    comp_cls = get_completion_class(shell)
    if comp_cls is None:  # pragma: no cover
        click.echo(f"Shell '{shell}' is not supported.", err=True)
        sys.exit(1)

    comp = comp_cls(cli, ctx_args, prog_name, complete_var)

    return comp.source()


def install_bash(
    cli: click.BaseCommand,
    ctx_args: Dict[str, Any],
    prog_name: str,
    complete_var: str,
    shell: str,
) -> Path:
    # Ref: https://github.com/scop/bash-completion#faq
    # It seems bash-completion is the official completion system for bash:
    # Ref: https://www.gnu.org/software/bash/manual/html_node/A-Programmable-Completion-Example.html
    # But installing in the locations from the docs doesn't seem to have effect
    completion_path = Path.home() / f".bash_completions/{prog_name}.sh"

    rc_path = Path.home() / ".bashrc"
    rc_path.parent.mkdir(parents=True, exist_ok=True)
    rc_content = ""
    if rc_path.is_file():
        rc_content = rc_path.read_text()

    completion_init_lines = [f"source {completion_path}"]
    for line in completion_init_lines:
        if line not in rc_content:  # pragma: no cover
            rc_content += f"\n{line}"

    rc_content += "\n"
    rc_path.write_text(rc_content)

    # Install completion
    completion_path.parent.mkdir(parents=True, exist_ok=True)
    script_content = get_completion_script(
        cli, ctx_args, prog_name, complete_var, shell
    )
    completion_path.write_text(script_content)

    return completion_path


def install_zsh(
    cli: click.BaseCommand,
    ctx_args: Dict[str, Any],
    prog_name: str,
    complete_var: str,
    shell: str,
) -> Path:
    # Set up Z shell and load `~/.zfunc``
    zshrc_path = Path.home() / ".zshrc"
    zshrc_path.parent.mkdir(parents=True, exist_ok=True)
    zshrc_content = ""
    if zshrc_path.is_file():
        zshrc_content = zshrc_path.read_text()

    completion_init_lines = [
        "autoload -Uz compinit",
        "compinit",
        "zstyle ':completion:*' menu select",
        "fpath+=~/.zfunc",
    ]
    for line in completion_init_lines:
        if line not in zshrc_content:  # pragma: no cover
            zshrc_content += f"\n{line}"

    zshrc_content += "\n"
    zshrc_path.write_text(zshrc_content)

    # Install completion under `~/.zfunc/``
    path_obj = Path.home() / f".zfunc/_{prog_name}"
    path_obj.parent.mkdir(parents=True, exist_ok=True)

    script_content = get_completion_script(
        cli, ctx_args, prog_name, complete_var, shell
    )
    path_obj.write_text(script_content)

    return path_obj


def install_fish(
    cli: click.BaseCommand,
    ctx_args: Dict[str, Any],
    prog_name: str,
    complete_var: str,
    shell: str,
) -> Path:
    path_obj = Path.home() / f".config/fish/completions/{prog_name}.fish"
    parent_dir: Path = path_obj.parent
    parent_dir.mkdir(parents=True, exist_ok=True)

    script_content = get_completion_script(
        cli, ctx_args, prog_name, complete_var, shell
    )
    path_obj.write_text(f"{script_content}\n")

    return path_obj


def install(
    cli: click.BaseCommand,
    ctx_args: Dict[str, Any],
    shell: Optional[str] = None,
    prog_name: Optional[str] = None,
    complete_var: Optional[str] = None,
) -> Tuple[str, Path]:
    prog_name = prog_name or click.get_current_context().find_root().info_name
    assert prog_name

    if complete_var is None:
        complete_var = "_{}_COMPLETE".format(prog_name.replace("-", "_").upper())

    test_disable_detection = os.getenv("_TYPER_COMPLETE_TEST_DISABLE_SHELL_DETECTION")
    if shell is None and shellingham is not None and not test_disable_detection:
        shell, _ = shellingham.detect_shell()

    if shell == BashComplete.name:
        installed_path = install_bash(cli, ctx_args, prog_name, complete_var, shell)
    elif shell == ZshComplete.name:
        installed_path = install_zsh(cli, ctx_args, prog_name, complete_var, shell)
    elif shell == FishComplete.name:
        installed_path = install_fish(cli, ctx_args, prog_name, complete_var, shell)
    else:
        click.echo(f"Shell '{shell}' is not supported.")
        raise click.exceptions.Exit(1)

    return shell, installed_path


def get_completion_inspect_parameters() -> Tuple[ParamMeta, ParamMeta]:
    test_disable_detection = os.getenv("_TYPER_COMPLETE_TEST_DISABLE_SHELL_DETECTION")
    if shellingham and not test_disable_detection:
        parameters = get_params_from_function(_install_completion_placeholder_function)
    else:
        parameters = get_params_from_function(
            _install_completion_no_auto_placeholder_function
        )
    install_param, show_param = parameters.values()
    return install_param, show_param


def install_callback(ctx: click.Context, param: click.Parameter, value: Any) -> Any:
    if not value or ctx.resilient_parsing:
        return value  # pragma no cover
    if isinstance(value, str):
        shell, path = install(ctx.command, {}, shell=value)
    else:
        shell, path = install(ctx.command, {})
    click.secho(f"{shell} completion installed at `{path}`", fg="green")
    click.echo("Completion will take effect once you restart the terminal")
    sys.exit(0)


def show_callback(ctx: click.Context, param: click.Parameter, value: Any) -> Any:
    if not value or ctx.resilient_parsing:
        return value  # pragma no cover
    prog_name = ctx.find_root().info_name
    assert prog_name
    complete_var = f"_{prog_name}_COMPLETE".replace("-", "_").upper()
    shell = ""
    test_disable_detection = os.getenv("_TYPER_COMPLETE_TEST_DISABLE_SHELL_DETECTION")
    if isinstance(value, str):
        shell = value
    elif shellingham and not test_disable_detection:
        shell, _ = shellingham.detect_shell()
    script_content = get_completion_script(
        ctx.command, {}, prog_name, complete_var, shell
    )
    click.echo(script_content)
    sys.exit(0)


# Create a fake command function to extract the completion parameters
def _install_completion_placeholder_function(
    install_completion: bool = Option(
        None,
        "--install-completion",
        is_flag=True,
        callback=install_callback,
        expose_value=False,
        help="Install completion for the current shell.",
    ),
    show_completion: bool = Option(
        None,
        "--show-completion",
        is_flag=True,
        callback=show_callback,
        expose_value=False,
        help="Show completion for the current shell, to copy it or customize the installation.",
    ),
) -> Any:
    pass  # pragma no cover


def _install_completion_no_auto_placeholder_function(
    install_completion: Shells = Option(
        None,
        callback=install_callback,
        expose_value=False,
        help="Install completion for the specified shell.",
    ),
    show_completion: Shells = Option(
        None,
        callback=show_callback,
        expose_value=False,
        help="Show completion for the specified shell, to copy it or customize the installation.",
    ),
) -> Any:
    pass  # pragma no cover


# Re-implement `shell_complete` to add error messages.
def shell_complete(
    cli: click.BaseCommand,
    ctx_args: Dict[str, Any],
    prog_name: str,
    complete_var: str,
    instruction: str,
) -> int:
    if "_" not in instruction:
        click.echo("Invalid completion instruction.", err=True)
        return 1

    shell, _, instruction = instruction.partition("_")
    comp_cls = get_completion_class(shell)
    if comp_cls is None:
        click.echo(f"Shell '{shell}' is not supported.", err=True)
        return 1

    comp = comp_cls(cli, ctx_args, prog_name, complete_var)

    if instruction == "source":
        click.echo(comp.source())
        return 0
    elif instruction == "complete":
        click.echo(comp.complete())
        return 0

    click.echo(f"Completion instruction '{instruction}' is not supported.", err=True)
    return 1



import os
import subprocess

from docs_src.commands.help import tutorial001 as mod


def test_completion_complete_subcommand_bash():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, " "],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TUTORIAL001.PY_COMPLETE": "bash_complete",
            "COMP_WORDS": "tutorial001.py del",
            "COMP_CWORD": "1",
            "_TYPER_COMPLETE_TESTING": "True",
        },
    )
    assert "plain,delete\n" "plain,delete-all\n" in result.stdout


def test_completion_complete_subcommand_bash_invalid():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, " "],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TUTORIAL001.PY_COMPLETE": "bash_complete",
            "COMP_WORDS": "tutorial001.py del",
            "COMP_CWORD": "42",
            "_TYPER_COMPLETE_TESTING": "True",
        },
    )
    assert (
        "plain,create\n"
        "plain,delete\n"
        "plain,delete-all\n"
        "plain,init\n" in result.stdout
    )


def test_completion_complete_subcommand_bash_argument():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, " "],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TUTORIAL001.PY_COMPLETE": "bash_complete",
            "COMP_WORDS": "tutorial001.py delete ",
            "COMP_CWORD": "2",
            "_TYPER_COMPLETE_TESTING": "True",
        },
    )
    assert "plain" not in result.stdout


def test_completion_complete_subcommand_zsh():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, " "],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TUTORIAL001.PY_COMPLETE": "zsh_complete",
            "COMP_WORDS": "tutorial001.py del",
            "COMP_CWORD": "1",
            "_TYPER_COMPLETE_TESTING": "True",
        },
    )
    assert (
        "plain\ndelete\nDelete a user with USERNAME.\n"
        "plain\ndelete-all\nDelete ALL users in the database.\n" in result.stdout
    )


def test_completion_complete_subcommand_fish():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, " "],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        env={
            **os.environ,
            "_TUTORIAL001.PY_COMPLETE": "fish_complete",
            "COMP_WORDS": "tutorial001.py del",
            "COMP_CWORD": "del",
            "_TYPER_COMPLETE_TESTING": "True",
        },
    )
    assert (
        "plain,delete\tDelete a user with USERNAME.\n"
        "plain,delete-all\tDelete ALL users in the database.\n" in result.stdout
    )



import typer_cloup as typer


def main(
    name: str, email: str = typer.Option(..., prompt=True, confirmation_prompt=True)
):
    typer.echo(f"Hello {name}, your email is {email}")


if __name__ == "__main__":
    typer.run(main)



from typing import Any, List

import typer_cloup as typer
from typer_cloup.param_types import SHELL_QUOTED_LIST
from typer_cloup.testing import CliRunner

runner = CliRunner()


def test_typer_unprocessed():
    app = typer.Typer()

    @app.command()
    def main(name: Any):
        assert isinstance(name, bytes) and name == b"\xa1"

    result = runner.invoke(app, [b"\xa1"])
    assert result.exit_code == 0


def test_shell_quoted_list():
    assert repr(SHELL_QUOTED_LIST) == "SHELL-QUOTED LIST"
    assert SHELL_QUOTED_LIST.convert(None, None, None) is None
    assert SHELL_QUOTED_LIST.convert(["foo", "bar"], None, None) == ["foo", "bar"]


def test_typer_shell_quoted_list():
    app = typer.Typer()

    @app.command()
    def main(
        names: List[str] = typer.Argument(..., shell_quoted_list=True, envvar="NAMES")
    ):
        for name in names:
            typer.echo(f"Hello {name}")

    result = runner.invoke(app, ["'Joe Bloggs' Camila"])
    assert "Hello Joe Bloggs" in result.stdout
    assert "Hello Camila" in result.stdout

    runner.env["NAMES"] = "'Joe Bloggs' Camila"
    result = runner.invoke(app)
    assert "Hello Joe Bloggs" in result.stdout
    assert "Hello Camila" in result.stdout



from typing import List

from click.shell_completion import CompletionItem

import typer_cloup as typer

valid_completion_items = [
    ("Camila", "The reader of books."),
    ("Carlos", "The writer of scripts."),
    ("Sebastian", "The type hints guy."),
]


def complete_name(ctx, param, incomplete):
    typer.echo(f"{ctx.args}", err=True)
    completion = []
    for name, help_text in valid_completion_items:
        if name.startswith(incomplete):
            completion.append(CompletionItem(name, help=help_text))
    return completion


app = typer.Typer()


@app.command()
def main(
    name: List[str] = typer.Option(
        ["World"], help="The name to say hi to.", shell_complete=complete_name
    )
):
    for n in name:
        typer.echo(f"Hello {n}")


if __name__ == "__main__":
    app()



import typer_cloup as typer

app = typer.Typer()


@app.command()
def main(name: str = typer.Option("World", help="The name to say hi to.")):
    typer.echo(f"Hello {name}")


if __name__ == "__main__":
    app()



import subprocess

import typer_cloup as typer
from docs_src.arguments.default import tutorial001 as mod
from typer_cloup.testing import CliRunner, columns_match

runner = CliRunner()

app = typer.Typer()
app.command()(mod.main)


def test_help():
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "[OPTIONS] [NAME]" in result.output
    assert "Arguments:" in result.output
    assert columns_match(result.output, "[NAME]", "[default: Wade Wilson]")


def test_call_no_arg():
    result = runner.invoke(app)
    assert result.exit_code == 0
    assert "Hello Wade Wilson" in result.output


def test_call_arg():
    result = runner.invoke(app, ["Camila"])
    assert result.exit_code == 0
    assert "Hello Camila" in result.output


def test_script():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, "--help"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
    )
    assert "Usage" in result.stdout



import subprocess

import typer_cloup as typer
from docs_src.parameter_types.index import tutorial001 as mod
from typer_cloup.testing import CliRunner

runner = CliRunner()

app = typer.Typer()
app.command()(mod.main)


def test_help():
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert "--age INTEGER" in result.output
    assert "--height-meters FLOAT" in result.output


def test_params():
    result = runner.invoke(
        app, ["Camila", "--age", "15", "--height-meters", "1.70", "--female"]
    )
    assert result.exit_code == 0
    assert "NAME is Camila, of type: <class 'str'>" in result.output
    assert "--age is 15, of type: <class 'int'>" in result.output
    assert "--height-meters is 1.7, of type: <class 'float'>" in result.output
    assert "--female is True, of type: <class 'bool'>" in result.output


def test_invalid():
    result = runner.invoke(app, ["Camila", "--age", "15.3"])
    assert result.exit_code != 0
    assert (
        "Error: Invalid value for '--age': '15.3' is not a valid integer"
        in result.output
    )


def test_script():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, "--help"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
    )
    assert "Usage" in result.stdout



import typer_cloup as typer


def main(
    id: int = typer.Argument(..., min=0, max=1000),
    rank: int = typer.Option(0, max=10, clamp=True),
    score: float = typer.Option(0, min=0, max=100, clamp=True),
):
    typer.echo(f"ID is {id}")
    typer.echo(f"--rank is {rank}")
    typer.echo(f"--score is {score}")


if __name__ == "__main__":
    typer.run(main)



import typer_cloup as typer
from typer_cloup.testing import CliRunner

TERMINAL_WIDTH = 800
ARG_HELP_TEXT = "This is an argument with a very long help text. " * 10

runner = CliRunner()

app = typer.Typer(
    context_settings=dict(
        max_content_width=TERMINAL_WIDTH,
        terminal_width=TERMINAL_WIDTH,
    )
)


@app.command()
def main(
    arg: str = typer.Argument(..., help=ARG_HELP_TEXT),
):
    pass  # pragma: no cover


def test_help_output():
    result = runner.invoke(app, ["--help"])
    assert result.exit_code == 0
    assert ARG_HELP_TEXT in result.output



import subprocess

import typer_cloup as typer
from docs_src.terminating import tutorial001 as mod
from typer_cloup.testing import CliRunner

runner = CliRunner()

app = typer.Typer()
app.command()(mod.main)


def test_cli():
    result = runner.invoke(app, ["Camila"])
    assert result.exit_code == 0
    assert "User created: Camila" in result.output
    assert "Notification sent for new user: Camila" in result.output


def test_existing():
    result = runner.invoke(app, ["rick"])
    assert result.exit_code == 0
    assert "The user already exists" in result.output
    assert "Notification sent for new user" not in result.output


def test_script():
    result = subprocess.run(
        ["coverage", "run", mod.__file__, "--help"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
    )
    assert "Usage" in result.stdout
